/* here are all options for the gp programs, stored in a static variable called
 * programs, of type programopt_s. Functions in this source file deal with
 * preparing the menu entries and creating a suitable options window if one of
 * these entries is called. Then a commandline is passed to function "execute"
 * from main.c */


#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <gtk/gtk.h>
#include "main.h"


/* two global variables. Sorry. */
entry_s *programs[ARKA_PROG_NUM_MAX] ;
int number_of_programs ;

/* function prototypes */

/* parses program file and returns an array with entries */
entry_s * parse_rc(FILE *in, int *i) ;
int append_items(opt_s *o, entry_s *e) ;

/* print an entry on stdout */
int entry_print(entry_s *e) ;
char *my_strdup(char *string) ;

/* -------------- some small, useful string funtions ------------- */

/* ANSI C doesn't have a strdup */
char *my_strdup(char *string) {
	int i ;
	char * res ;

	i = strlen(string) ;
	res = malloc(i + 1) ;
	strcpy(res, string) ;
	return res ;
}


/* exchange all occurencies of 'from' in 'string' by 'to' */
char *my_strsubst(char *string, char *from, char *to) {
	char *tmp, *t, *p, *res ;

	res = NULL ;
	tmp = my_strdup(string) ;
	t = tmp ;

	if(!string || !from || !to) return NULL ;

	while( (p = strstr(t, from)) ) {
		*p = '\0' ;
		if(!res) {
			res = malloc(strlen(t) + strlen(to) + 1) ;
			*res = '\0' ;
		} else res = realloc(res, strlen(res) + strlen(t) + strlen(to) + 1) ;
		strcat(res, t) ;
		strcat(res, to) ;
		t = p + strlen(from) ;
	}

	if(!res) {
		res = malloc(strlen(t) + 1) ;
		strcpy(res, t) ;
	} else {
		res = realloc(res, strlen(res) + strlen(t) + 1) ;
		strcat(res, t) ;
	}

	free(tmp) ;
	return res ;
}



/* ----------------------- the "new program" dialog ---------------------------- */
/* the dialog used to create new programs specs and related functions */


/* when the "next" button is pressed (not the first time, though) */
void programs_create_next(GtkWidget *kontr, gpointer dane) {
	data_s *data ;

	data = (data_s *) dane ;
	gtk_notebook_next_page(GTK_NOTEBOOK(data->window1)) ;
}


/* when the "prev" button is pressed */
void programs_create_back(GtkWidget *kontr, gpointer dane) {
	data_s *data ;

	data = (data_s *) dane ;
	gtk_notebook_prev_page(GTK_NOTEBOOK(data->window1)) ;
}


/* converts type description into type */
int option_string_to_type(char *type) {
	int res ;

	if(!strcasecmp(type, "boolean")) res = BOOLEAN ;
	if(!strcasecmp(type, "integer")) res = INT ;
	if(!strcasecmp(type, "file")) res = FILENAME ;
	if(!strcasecmp(type, "float") || !strcasecmp(type, "double")) res = DOUBLE ;
	if(!strcasecmp(type, "string")) res = STRING ;
	if(!strcasecmp(type, "hidden")) res = HIDDEN ;
	return res ;
}


/* returns a "[program] [options]" string from an entry_s structure */
char *option_usage_string(entry_s *e) {
	int l = 0, i ;
	entry_option_s *o ;
	char *res ;

	/* reserving memory */
	l += strlen(e->name) ;
	for(i = 0 ; i < e->nopts ; i++)
		l += strlen(e->options[i].designator) + 15 ;

	res = malloc(l) ;

	strcpy(res, e->name) ;

	/* adding options to the string */
	for(i = 0 ; i < e->nopts ; i++) {
		o = &e->options[i] ;

		if(o->type == HIDDEN) continue ;

		if(o->required) strcat(res, " <") ;
		else strcat(res, " [") ;

		if(o->designator) strcat(res, o->designator) ;

		switch(o->type){
			case BOOLEAN:
				break ;
			case INT:
				strcat(res, " int_value") ;
				break ;
			case DOUBLE:
				strcat(res, " float_value") ;
				break ;
			case FILENAME:
				strcat(res, " file") ;
				break ;
			case STRING:
				strcat(res, " string") ;
				break ;
			default:
				break ;
		}
		if(o->required) strcat(res, " >") ;
		else strcat(res, " ]") ;
	}

	return res ;
}


/* saves a program spec in a format which can be read by arka */
int print_program_tofile(FILE *fp, entry_s *p) {
	int i ;
	entry_option_s *o ;

	fprintf(fp, "start(program)(%s)(%s)(%s)", p->name, p->entry, p->branch) ;
	fprintf(fp, "(%s)\n", p->short_description) ;
	fprintf(fp, "\tstart(options)\n") ;

	/* saving options */
	for(i = 0 ; i < p->nopts ; i++) {
		o = &p->options[i] ;
		if(o->designator) fprintf(fp, "\t\t(%s)", o->designator) ;
		else fprintf(fp, "\t\t()") ;

		switch(o->type) {
			case INT:
				fprintf(fp, "(int)") ;
				break ;
			case DOUBLE:
				fprintf(fp, "(float)") ;
				break ;
			case STRING:
				fprintf(fp, "(string)") ;
				break ;
			case HIDDEN:
				fprintf(fp, "(invisible)") ;
				break ;
			case FILENAME:
				fprintf(fp, "(file)") ;
				break ;
			case BOOLEAN:
			default:
				fprintf(fp, "()") ;
				break ;
		}

		if(o->defvalue) fprintf(fp, "(%s)", o->defvalue) ;
		else fprintf(fp, "()") ;
		if(o->description) fprintf(fp, "(%s)", o->description) ;
		else fprintf(fp, "()") ;
		fprintf(fp, "\n") ;
	}

	fprintf(fp, "\tend(options)\n") ;
	fprintf(fp, "\tstart(description)\n") ;
	if(p->long_description) fprintf(fp, "%s", p->long_description) ;
	fprintf(fp, "\tend(description)\n") ;
	fprintf(fp, "end(program)\n\n") ;

	return EXIT_SUCCESS ;
}


/* used in the dialog to get the program description from widgets */
int extract_program_from_entries(entry_s *p, GtkWidget** e) {

	Warn("extracting program data") ;
	p->name = gtk_editable_get_chars(GTK_EDITABLE(e[0]), 0, -1) ;

	/* only allow to proceede if program name defined */
	if(strlen(p->name) == 0) { 
		komunikat("You have to enter at least the p name") ;
		g_free(p->name) ; 
		return EXIT_FAILURE ; 
	} else Warn("name is %s", p->name) ;

	p->entry = gtk_editable_get_chars(GTK_EDITABLE(e[1]), 0, -1) ;

	if(p->entry && strlen(p->entry) == 0) {
		g_free(p->entry) ;
		p->entry = my_strdup(p->name) ;
	}

	p->short_description = gtk_editable_get_chars(GTK_EDITABLE(e[2]), 0, -1) ;
	p->branch = gtk_editable_get_chars(GTK_EDITABLE(e[3]), 0, -1) ;

	if(e[4]) 
		p->long_description = gtk_editable_get_chars(GTK_EDITABLE(e[4]), 0, -1) ; 

	return EXIT_SUCCESS ;
}


/* called by the `save' button -- saves the options, adds program entry */
void programs_save(GtkWidget *kontr, gpointer dane) {
	GtkWidget *okno, **entries ;
	entry_option_s * dialog ;
	entry_s * p ;
	data_s *data ;
	opt_s *main_o ;
	FILE *fp ;
	char *filename ;

	/* casting provided data into apriopriate types */
	data = (data_s *) dane ;
	main_o = (opt_s *) data->data ;

	p = (entry_s*) data->data1 ;
	dialog = (entry_option_s*) data->data2 ;
	entries = (GtkWidget **) data->data3 ;
	okno = data->window ;

	if(extract_program_from_entries(p, entries) == EXIT_FAILURE) return ;

	/* saving the data to file */

	/* getting rc directory */
	if(! (filename = my_strdup(getenv("HOME"))) ) {
		fp = NULL ;
	} else {
		filename = realloc(filename, strlen(filename) + strlen(ARKA_DIR) + 20) ;
		strcat(filename, ARKA_DIR) ;
		strcat(filename, "/programs") ;
		Warn("Saving spec for program %s to file %s", p->name, filename) ;
		fp = fopen(filename, "a") ;
		free(filename) ;
	}

	if(!fp) {
		komunikat("I cannot open $HOME/%s/programs file for writing.\n"
			"I will print the program specification to the console:\n"
			"That way, you can cut & paste the spec and save it to a file", ARKA_DIR) ;
		fp = stdout ;
	} 

	print_program_tofile(fp, p) ;
	if(fp && fp != stdout) fclose(fp) ;

	/* appending entry */
	if(number_of_programs == ARKA_PROG_NUM_MAX) {
		komunikat("Sorry. Arka can now accept only %i programs", ARKA_PROG_NUM_MAX) ;
	} else {
		p->number = number_of_programs ;
		number_of_programs++ ;
		append_items(main_o, p) ;
		programs[p->number] = p ;
	}

	gtk_widget_destroy(okno) ;
	gtk_main_quit() ;
}


/* called when the "add" button is pressed */
void programs_add_option(GtkWidget *kontr, gpointer dane) {
	GtkWidget **entries ;
	entry_option_s * dialog, *o ;
	entry_s * program ;
	data_s *data ;
	char *tmp ;
	int i ;

	/* casting provided data into apriopriate types */
	data = (data_s *) dane ;

	program = (entry_s*) data->data1 ;
	dialog = (entry_option_s*) data->data2 ;
	entries = (GtkWidget **) data->data3 ;

	tmp = gtk_editable_get_chars(GTK_EDITABLE(dialog[1].widget), 0, -1) ;
	if(!tmp || strlen(tmp) == 0) {
		free(tmp) ;
		komunikat("Please enter the option description") ;
		return ;
	}

	/* reserving space for the next option */
	if(program->nopts == 0) program->options = malloc(sizeof(*program->options)) ;
	else 
		program->options = realloc(program->options, (program->nopts + 1) *
		sizeof(*program->options)) ;
	
	if(!program->options) {
		komunikat("Could not allocate enough memory!") ;
		program->nopts = 0 ;
		return ;
	}

	program->nopts++ ;

	/* working on the last element */
	o = &program->options[program->nopts - 1] ;

	/* getting parameters from the widgets */
	o->set = FALSE ;
	o->description = tmp ;
	o->designator = gtk_editable_get_chars(GTK_EDITABLE(dialog[2].widget), 0, -1) ;
	o->defvalue = gtk_editable_get_chars(GTK_EDITABLE(dialog[3].widget), 0, -1) ;
	tmp = gtk_editable_get_chars(GTK_EDITABLE(GTK_COMBO(dialog[4].widget)->entry), 0, -1) ;
	o->type = option_string_to_type(tmp) ;
	free(tmp) ;

	tmp = gtk_editable_get_chars(GTK_EDITABLE(GTK_COMBO(dialog[5].widget)->entry), 0, -1) ;
	if(strcmp(tmp, "required") == 0) o->required = TRUE ;
	else o->required = FALSE ;

	tmp = option_usage_string(program) ;
	gtk_label_set_text(GTK_LABEL(dialog[0].widget), tmp) ;
	free(tmp) ;
		
	/* deleting entries */
	gtk_editable_delete_text(GTK_EDITABLE(dialog[1].widget), 0, -1) ;
	gtk_editable_delete_text(GTK_EDITABLE(dialog[2].widget), 0, -1) ;
	gtk_editable_delete_text(GTK_EDITABLE(dialog[3].widget), 0, -1) ;
}


/* when the "next" button is pressed -- create a new panel for options */
void programs_create_options(GtkWidget *kontr, gpointer dane) {
	GtkWidget *new, *label, *ok, *next, *cancel, *ypole, *hpole, *nbook,
		*item, *sep, **entries, *okno ;
	data_s *data ;
	entry_option_s *dialog ;
	entry_s *program ;
	char *name, *tmp,
	*lista[] = { "boolean", "integer", "float", "string", "file", "hidden" },
	*lista2[] = { "optional", "required" } ;

	data = (data_s *) dane ;

	okno = data->window ;
	nbook = data->window1 ;
	next = data->window2 ;
	hpole = data->window4 ;

	program = (entry_s*) data->data1 ;
	dialog = (entry_option_s*) data->data2 ;
	entries = (GtkWidget **) data->data3 ;

	if(extract_program_from_entries(program, entries) == EXIT_FAILURE) return ;
	Warn("programs_create_options: program name %s", program->name) ;

	/* destroying the old function of next */
	gtk_widget_destroy(GTK_WIDGET(next)) ;

	ok = create_button(hpole, programs_save, "Save", data) ;
	next = create_button(hpole, programs_create_next, "Next >", data) ;

	ypole = gtk_vbox_new(FALSE, 5) ;
	label = gtk_label_new("Add options") ;

	/* the new notebook page with the options dialog */
	gtk_notebook_append_page(GTK_NOTEBOOK(nbook), ypole, label) ;

	create_label(ypole, "Here you can add options to the program\n -- one by one") ;
	create_separator(ypole) ;
	create_label(ypole, "Current commandline:") ;
	dialog[0].widget = create_label(ypole, program->name) ;

	create_separator(ypole) ;

	/* creating widgets */
	dialog[2].widget = create_option_widget(ypole, STRING, "designator") ;
	dialog[1].widget = create_option_widget(ypole, STRING, "description") ;
	dialog[3].widget = create_option_widget(ypole, STRING, "default value ") ;
	dialog[4].widget = create_combo(ypole, "type", lista, 6) ;
	dialog[5].widget = create_combo(ypole, "required / not required", lista2, 2) ;

	/* making buttons */
	hpole = create_hbox(ypole) ;
	next = create_button(hpole, programs_create_back, "< Previous", data) ;
	next = create_button(hpole, zakoncz_dialog, "Cancel", okno) ;
	ok = create_button(hpole, programs_add_option, "Add Option", data) ;
	next = create_button(hpole, programs_create_next, "Next >", data) ;

	/* notebook page for the long description of the program */
	ypole = gtk_vbox_new(FALSE, 5) ;
	label = gtk_label_new("Long description") ;
	gtk_notebook_append_page(GTK_NOTEBOOK(nbook), ypole, label) ;

	create_label(ypole, "Here is place for a longer description of your program") ;
	create_separator(ypole) ;

	entries[4] = create_text(ypole, NULL, TRUE) ;

	create_separator(ypole) ;

	hpole = create_hbox(ypole) ;
	next = create_button(hpole, programs_create_back, "< Previous", data) ;
	next = create_button(hpole, zakoncz_dialog, "Cancel", okno) ;


	gtk_widget_show_all(okno) ;
	gtk_notebook_next_page(GTK_NOTEBOOK(nbook)) ;
}


/* creating the "new program" dialog */
void programs_create(gpointer dane, guint signal, GtkWidget *kontrolka) {
	GtkWidget *label, *next, *cancel, *ypole, *hpole, *nbook,
		*sep, *entries[10], *okno ;
	entry_option_s dialog[10] ;
	entry_s *program ;
	opt_s *o ;
	data_s data ;
	int i ;

	o = (opt_s *) dane ;

	for(i = 0 ; i < 10 ; i++) entries[i] = NULL ;

	program = malloc(sizeof(*program)) ;
	program->name = NULL ;
	program->entry = NULL ;
	program->options = NULL ;
	program->nopts = 0 ;
	program->short_description = NULL ;
	program->long_description = NULL ;
	program->branch = NULL ;

	data.data = o ;
	data.data1 = program ;
	data.data2 = dialog ;

	okno = gtk_window_new(GTK_WINDOW_DIALOG) ;
	gtk_window_set_title(GTK_WINDOW(okno), "New program entry") ;
	gtk_signal_connect(GTK_OBJECT(okno), "destroy", GTK_SIGNAL_FUNC(gtk_main_quit), okno) ;

	Warn("programs_create") ;
	nbook = gtk_notebook_new() ;

	ypole = gtk_vbox_new(FALSE, 5) ;

	create_separator(ypole) ;

	entries[0] = create_option_widget(ypole, STRING, "program name") ;
	entries[1] = create_option_widget(ypole, STRING, "menu entry") ;
	entries[2] = create_option_widget(ypole, STRING, "short description") ;
	entries[3] = create_option_widget(ypole, STRING, "program branch") ;

	create_separator(ypole) ;
	hpole = create_hbox(ypole) ;

	cancel = create_button(hpole, zakoncz_dialog, "Cancel", okno) ;
	next = create_button(hpole, programs_create_options, "Next >", &data) ;

	/* some variables we have to remember */
	data.window = okno ;
	data.window1 = nbook ;
	data.window2 = next ;
	data.window4 = hpole ;

	data.data3 = entries ;

	label = gtk_label_new("New program") ;
	gtk_notebook_append_page(GTK_NOTEBOOK(nbook), ypole, label) ;

	gtk_container_add(GTK_CONTAINER(okno), nbook) ;
	gtk_widget_show_all(okno) ;

	gtk_main() ;
}


/* ------------------------ execution of the programs ---------------------- */
/* the dialog used to start program execution and related functions */

/* catches the signal from menu and executes the program */
void program_select(gpointer dane, guint signal, GtkWidget *kontrolka) {
	opt_s *o ;
	char *cmdline ;

	o = (opt_s *) dane ;
	cmdline = set_options(signal, o) ; /* do the dialog stuff */

	Warn("program_select: got: %s\n from set_options", cmdline) ;

	if(!cmdline) return ;

	Warn("program_select: executing\n%s", cmdline) ;
	o->command_history = g_list_prepend(o->command_history, cmdline) ;

	execute(cmdline, o) ;

}

/* given an entry, append it to the main menu */
int append_items(opt_s *o, entry_s *e) {
	GtkItemFactoryEntry element = {NULL, NULL,	program_select,	0} ;
	char *t ;

	/* preparing item_factory item */
	t = malloc(15 + strlen(e->branch) + strlen(e->entry)) ;
	if(e->branch && strlen(e->branch) > 0)
		sprintf(t, "/Programs/%s/%s", e->branch, e->entry) ;
	else
		sprintf(t, "/Programs/%s", e->entry) ;

	element.path = t ;
	element.callback_action = e->number ;
	Warn("path: %s entry: %s", t, e->entry) ;

	/* creating menu entry */
	gtk_item_factory_create_items(o->fabryka, 1, &element, o) ;
	free(t) ;
	return EXIT_SUCCESS ;

}


/* this function sits in between programs_append() and functions actually
 * parsing the description file. It records the descriptions and makes an
 * apriopriate menu entry. */
int parse_and_append(opt_s *o, FILE* fp, int start) {
	int i, line = 0 ;
	entry_s *e ;

	i = start ; 

	while( (e = parse_rc(fp, &line)) != NULL ) {

		if( start >= ARKA_PROG_NUM_MAX) {
			Warn("Sorry. Arka can accept only %i programs right now.\n"
				"This should be easy enough to change, but right now I am too lazy to do it.\n",
			ARKA_PROG_NUM_MAX) ;
			return i ;
		}

		e->number = i ;

		append_items(o, e) ;

		programs[i] = e ;
		i++ ;
	}
	fclose(fp) ;

	return i ;

}


/* appends programs to the main program menu bar */
void programs_append(opt_s *o) {
	GtkItemFactoryEntry sep = {"/Programs/sep2",	NULL,	NULL,	0, "<Separator>"}, 
		new = {"/Programs/_New...", NULL, programs_create, 0} ;
	int i = 0  ;
	FILE *fp ;
	char *rcfile ;

	/* reading program description from the global file */
	if( !(fp = fopen(ARKA_PROGRAMS, "r")) ) { 
		Warn("File %s could not be opened", ARKA_PROGRAMS) ;
	} else {
		Warn("reading %s", ARKA_PROGRAMS) ;
		i = parse_and_append(o, fp, i) ;
	}

	if( !(rcfile = my_strdup(getenv("HOME")) ) ) {
		Warn("$HOME undefined.") ;
		return ;
	}

	/* reading program description from the local file */
	rcfile = realloc(rcfile, strlen(rcfile) + strlen(ARKA_DIR) + 20) ;
	strcat(rcfile, ARKA_DIR) ;
	strcat(rcfile, "/programs") ;

	if( !(fp = fopen(rcfile, "r")) ) { 
		Warn("File %s could not be opened", rcfile) ;
	} else {
		Warn("reading %s", rcfile) ;
		i = parse_and_append(o, fp, i) ;
	}

	number_of_programs = i ;

	gtk_item_factory_create_items(o->fabryka, 1, &sep, o) ;
	gtk_item_factory_create_items(o->fabryka, 1, &new, o) ;
	Warn("%i programs loaded", number_of_programs) ;
	return ;
}


/* ----------------------- program dialog - related functions ----------------- */

/* this is called by the "Browse..." button in the main options dialog */
void browse_files(GtkWidget *kontr, gpointer dane) {
	opt_s *main_o ;
	entry_option_s *o ;

	o = (entry_option_s *) dane ;
	main_o = o->main_opts ;

	if(o->strvalue) free(o->strvalue) ;
	o->strvalue = file_get_name(main_o, NULL) ;
	gtk_entry_set_text(GTK_ENTRY(o->widget), o->strvalue) ;

}


/* clicking on "check" button calls this function, which records the change */
void toggle_optstat(GtkWidget *kontr, gpointer dane) {
	int *ip ;

	ip = (int *) dane ;
	*ip = ! *ip ;
}


/* called if "OK" in the option dialog is pressed. Computates the option line */
void close_options(GtkWidget *kontr, gpointer dane) {
	data_s *d ;
	int i ;
	char *res, tmp[100] ;
	entry_s *p ;
	entry_option_s *o ;
	
	/* casting data-to-be-processed into apriopriate types */
	d = (data_s *) dane ;
	p = (entry_s *) d->data1 ;

	res = malloc(2000) ; /* yeah I know this is evil */
	Warn("close_options: chosen program is %s", p->name) ;
	sprintf(res, "%s ", p->name) ;

	/* for each option... */
	for(i = 0 ; res && i < p->nopts ; i++) {

		o = &p->options[i] ;

		switch(o->type){
			case BOOLEAN:
				if(o->set) {
					sprintf(tmp, "%s ", o->designator) ;
					strcat(res, tmp) ;
				}
				break ;

			case INT:
				if(o->set) {
					o->intvalue = gtk_spin_button_get_value_as_int(GTK_SPIN_BUTTON(o->widget)) ;
					sprintf(tmp, "%s %i ", o->designator, o->intvalue) ;
					strcat(res, tmp) ;
				}
				break ;

			case DOUBLE:
				if(o->set) {
					o->doublevalue = 
						gtk_spin_button_get_value_as_float(GTK_SPIN_BUTTON(o->widget)) ;
					sprintf(tmp, "%s %f ", o->designator, o->doublevalue) ;
					strcat(res, tmp) ;
				}
				break ;

			case STRING:
			case FILENAME:
				/* getting the user data */
				if(o->strvalue) free(o->strvalue) ;
				o->strvalue = gtk_editable_get_chars(GTK_EDITABLE(o->widget), 0, -1) ;
				if(o->strvalue && strlen(o->strvalue) > 0) o->set = TRUE ;
				else o->set = FALSE ;

				/* if entry field empty, the option is not added */
				if(o->set){
					sprintf(tmp, "%s %s ", o->designator, o->strvalue) ;
					strcat(res, tmp) ;
				}

				break ;

			case HIDDEN:
				sprintf(tmp, "%s %s ", o->designator, o->defvalue) ;
				strcat(res, tmp) ;
			break ;

			default:
			break ;
		}
	}

	gtk_widget_destroy(d->window1) ;
	gtk_main_quit() ;
	d->data = res ;
}


/* used by set_options. Produces a widget for a single option */
GtkWidget * set_widget_for_option(entry_option_s *o, GtkWidget *parentbox) {
	GtkWidget *hpole, *res, *tmp, *optlab ;
	GtkAdjustment*adj ; 
	res = NULL ;

	hpole = gtk_hbox_new(FALSE, 5) ;
	gtk_box_pack_start(GTK_BOX(parentbox), hpole, FALSE, FALSE, 2) ;

	switch(o->type){

		case BOOLEAN:
			/* just a simple check button */
			res = gtk_check_button_new_with_label(o->description) ;
			gtk_box_pack_start(GTK_BOX(hpole), res, FALSE, FALSE, 5) ;
			gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(res), o->set) ;
			gtk_signal_connect(GTK_OBJECT(res), "toggled",
				GTK_SIGNAL_FUNC(toggle_optstat), &o->set) ;
			break ;

		case INT:
		case DOUBLE:
			/* checkbox */
			if(!o->required) {
				tmp = gtk_check_button_new() ;
				gtk_box_pack_start(GTK_BOX(hpole), tmp, FALSE, FALSE, 5) ;
				gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(tmp), o->set) ;
				gtk_signal_connect(GTK_OBJECT(tmp), "toggled",
					GTK_SIGNAL_FUNC(toggle_optstat), &o->set) ;
			}

			/* spin button */
			adj = (GtkAdjustment *) gtk_adjustment_new(0, INT_MIN, INT_MAX, 1, 10, 10) ;

			if(o->type == INT) res = gtk_spin_button_new(adj, 1, 0) ;
			else res = gtk_spin_button_new(adj, 1, 2) ;

			gtk_box_pack_start(GTK_BOX(hpole), res, FALSE, FALSE, 5) ;
			if(o->set) gtk_spin_button_set_value(GTK_SPIN_BUTTON(res), o->intvalue) ;
			gtk_spin_button_set_numeric(GTK_SPIN_BUTTON(res), FALSE) ;

			/* label */
			optlab = gtk_label_new(o->description) ;
			gtk_box_pack_start(GTK_BOX(hpole), optlab, FALSE, FALSE, 5) ;
			gtk_label_set_justify(GTK_LABEL(optlab), GTK_JUSTIFY_LEFT) ;
			gtk_label_set_line_wrap(GTK_LABEL(optlab), TRUE) ;
			break ;

		case STRING:
		case FILENAME:
			/* option entry */
			res = gtk_entry_new() ;
			gtk_box_pack_start(GTK_BOX(hpole), res, FALSE, FALSE, 5) ;
			if(o->strvalue) gtk_entry_set_text(GTK_ENTRY(res), o->strvalue) ;

			/* when clicked, we go through the directory */
			if(o->type == FILENAME) {
				tmp = gtk_button_new_with_label("Browse...") ;
				gtk_box_pack_start(GTK_BOX(hpole), tmp, FALSE, FALSE, 5) ;
				gtk_signal_connect(GTK_OBJECT(tmp), "clicked",
					GTK_SIGNAL_FUNC(browse_files), o) ;
			}

			/* description of this option */
			optlab = gtk_label_new(o->description) ;
			gtk_box_pack_start(GTK_BOX(hpole), optlab, FALSE, FALSE, 5) ;
			gtk_label_set_justify(GTK_LABEL(optlab), GTK_JUSTIFY_LEFT) ;
			gtk_label_set_line_wrap(GTK_LABEL(optlab), TRUE) ;

			break ;
		case HIDDEN:
			break ;
		default:
			Warn("unsupported option!") ;
			break ;
	}

	return res ;
}

/* produces a dialog for choosing option for a given program and returns the
 * programs option string */
char* set_options(int wh, opt_s *main_opts) {
	GtkWidget *tmp, *sep, *label, *okno, *ypole, *hpole, *man, *ok, *cancel ;
	GtkWidget *notebook ;
	int i, j ;
	char *res = NULL ;
	data_s d ;
	entry_option_s *o ;

	/* this structure will hold the data passed to close_options() */
	d.data = NULL ;
	d.data1 = programs[wh] ;

	Warn("set_options: selection %i", wh) ;
	Warn("name: %s", programs[wh]->name) ;

	/* main option window */
	okno = gtk_window_new(GTK_WINDOW_DIALOG) ;
	gtk_window_set_title(GTK_WINDOW(okno), programs[wh]->name) ;
	gtk_signal_connect(GTK_OBJECT(okno), "destroy",
		GTK_SIGNAL_FUNC(gtk_main_quit), okno) ;

	d.window1 = okno ;

	label = gtk_label_new("Main") ;
	ypole = gtk_vbox_new(FALSE, 5) ;
	notebook = gtk_notebook_new() ;
	gtk_notebook_append_page(GTK_NOTEBOOK(notebook), ypole, label) ;
	gtk_container_add(GTK_CONTAINER(okno), notebook) ;

	if(programs[wh]->long_description) {
		label = gtk_label_new("More..") ;
		tmp = gtk_label_new(programs[wh]->long_description) ;
		gtk_label_set_justify(GTK_LABEL(tmp), GTK_JUSTIFY_LEFT) ;
		gtk_notebook_append_page(GTK_NOTEBOOK(notebook), tmp, label) ;
	}

	/* brief program description */
	label = gtk_label_new(programs[wh]->short_description) ;
	gtk_box_pack_start(GTK_BOX(ypole), label, FALSE, FALSE, 2) ;
	gtk_label_set_line_wrap(GTK_LABEL(label), TRUE) ;

	sep = gtk_hseparator_new() ;
	gtk_box_pack_start(GTK_BOX(ypole), sep, FALSE, FALSE, 2) ;

	/* for each option, we create a widget stored in opsent and record a state in
	 * optstat */
	for(i = 0, j = 0 ; i < programs[wh]->nopts ; i++, j++) {
		o = &programs[wh]->options[i] ;

		/* we need to pass the main_opts pointer to file_get_name */
		o->main_opts = main_opts ;

		/* create an apriopriate widget for the option */
		o->widget = set_widget_for_option(o, ypole) ;
	}

	/* separator + OK + cancel buttons */
	sep = gtk_hseparator_new() ;
	gtk_box_pack_start(GTK_BOX(ypole), sep, FALSE, FALSE, 5) ;

	hpole = gtk_hbox_new(FALSE, 5) ;
	gtk_box_pack_start(GTK_BOX(ypole), hpole, FALSE, FALSE, 5) ;

	/* ok and cancel buttons */
	ok = gtk_button_new_with_label("OK") ;
	gtk_box_pack_start(GTK_BOX(hpole), ok, FALSE, FALSE, 10) ;
	gtk_signal_connect(GTK_OBJECT(ok), "clicked",
		GTK_SIGNAL_FUNC(close_options), &d) ;

	cancel = gtk_button_new_with_label("Cancel") ;
	gtk_box_pack_start(GTK_BOX(hpole), cancel, FALSE, FALSE, 10) ;
	gtk_signal_connect(GTK_OBJECT(cancel), "clicked",
		GTK_SIGNAL_FUNC(zakoncz_dialog), okno) ;

	man = gtk_button_new_with_label("Man page") ;
	gtk_box_pack_start(GTK_BOX(hpole), man, FALSE, FALSE, 10) ;
	gtk_signal_connect(GTK_OBJECT(man), "clicked",
		GTK_SIGNAL_FUNC(help_display_man), programs[wh]->name) ;

	GTK_WIDGET_SET_FLAGS(ok, GTK_CAN_DEFAULT) ;
	gtk_widget_grab_default(ok) ;
	gtk_widget_show_all(okno) ;
	gtk_main() ;

	/* the resulting command line gets stored by close_options() in d.data */
	res = (char*) d.data ;
	if(res) printf("command line: %s \n", res) ;
	return res ;
}


/* ---------- f. related to parsing the config files ------------------ */
/* mostly text processing, no gtk at all */

/* prints an entry */
int entry_print(entry_s *e) {
	int i ;
	entry_option_s *o ;

	Warn("printing entry") ;
	printf("%s - %s\n\n", e->entry, e->short_description) ;
	printf("Usage:\n\t%s ", e->name) ;
	for(i = 0 ; i < e->nopts ; i++) {
		o = &e->options[i] ;
		if(o->type == HIDDEN) continue ;
		if(o->required) printf("<") ;
		else printf("[") ;
		if(o->designator) {
			printf("%s", o->designator) ;
			if(o->type) printf(" ") ;
		}
		switch(o->type){
			case BOOLEAN:
				break ;
			case INT:
				printf("int_value") ;
				break ;
			case DOUBLE:
				printf("float_value") ;
				break ;
			case FILENAME:
				printf("file") ;
				break ;
			case STRING:
				printf("string") ;
				break ;
			default:
				break ;
		}
		if(o->required) printf("> ") ;
		else printf("] ") ;
	}
	printf("\n") ;


	for(i = 0 ; i < e->nopts ; i++) {
		if(!i) printf("Options:\n") ;
		o = &e->options[i] ;
		if(o->type == 4) continue ;
		printf("\t") ;
		if(o->designator) printf("%s ", o->designator) ;
		switch(o->type){
			case 0:
				break ;
			case 1:
				printf("int_value ") ;
				break ;
			case 2:
				printf("float_value ") ;
				break ;
			case 3:
				printf("file ") ;
				break ;
			default:
				break ;
		}

		if(o->description) printf(": %s ", o->description) ;
		if(o->defvalue) printf(" (default: %s) ", o->defvalue) ;
		printf("\n") ;
	}
	printf("\n") ;

	if(e->long_description) printf("Description:\n\n%s\n", e->long_description) ;

	return EXIT_SUCCESS ;
}


/* parses the line starting the program section, which contains program
 * name and the short description */
int parse_program(char *line, entry_s *entries) {
	char *p, *q ;

	p = strchr(line, '(') ;

	/* reading program name; if no name is given, do not do anything */
	if(!p) return EXIT_FAILURE ;

	/* initializing the new entry_s structure */
	entries->short_description = NULL ;
	entries->branch = NULL ;
	entries->long_description = NULL ;
	entries->options = NULL ;
	entries->nopts = 0 ;

	/* start the actual parsing */
	p++ ;
	q = strchr(line, ')') ;
	if(q) { *q = '\0' ; q++ ; }
	entries->name = my_strdup(p) ;

	p = strchr(q, '(') ;
	if(!p) {
		entries->entry = my_strdup(entries->name) ;
		return EXIT_SUCCESS ;
	}

	/* reading menu entry */
	p++ ;
	q = strchr(p, ')') ;
	if(q) { *q = '\0' ; q++ ; }

	if(strlen(p) > 0) entries->entry = my_strdup(p) ;
	else entries->entry = my_strdup(entries->name) ;

	p = strchr(q, '(') ;
	if(!p) return EXIT_SUCCESS ;

	/* reading branch information */
	p++ ;
	q = strchr(p, ')') ;
	if(q) { *q = '\0' ; q++ ; }  
	entries->branch = my_strdup(p) ;

	/* reading short description */
	p = strchr(q, '(') ;
	if(!p) {
		entries->short_description = my_strdup("no description") ;
		return EXIT_SUCCESS ;
	}

	p++ ;
	q = strchr(p, ')') ;
	if(q) *q = '\0' ; 
	entries->short_description = my_strdup(p) ;

	return EXIT_SUCCESS ;
}


/* adds a line to the long description of the program, allocating memory if
 * needed */
int parse_description(char *line, entry_s *e) {
	int i ;

	if(!e->long_description) {
		e->long_description = malloc(strlen(line) + 1) ;
		*e->long_description = '\0' ;
	} else {
		i = strlen(line) + strlen(e->long_description) + 1 ;
		e->long_description = realloc(e->long_description, i) ;
	}

	if(!e->long_description) return EXIT_FAILURE ;
	strcat(e->long_description, line) ;

	return EXIT_SUCCESS ;
}


/* parses an option line and stores it in the apropriate structure */
/* see the sample rcfile for more information */
int parse_option(char *line, entry_s *e, int i) {
	entry_option_s * o ;
	char *p, *q ;

	p = strchr(line, '(') ;
	if(!p) return EXIT_FAILURE ;

	e->nopts++ ;

	/* reserving necessary memory; options are stored in an array */
	if(!e->options) e->options = malloc(sizeof(*(e->options))) ;
	else e->options = realloc(e->options, 
		e->nopts * sizeof(*(e->options))) ;

	/* current option */
	o = &e->options[e->nopts - 1] ;

	/* default settings. From now on, we can exit anytime */
	o->designator = NULL ;
	o->description = NULL ;
	o->defvalue = NULL ;
	o->required = FALSE ;
	o->type = BOOLEAN ;
	o->set = FALSE ;

	o->strvalue = NULL ;
	o->intvalue = 0 ;
	o->doublevalue = 0.0 ;

	/* reading designator, that is e.g. 'h' for option '-h' */
	q = strchr(p, ')') ;
	p++ ;

	if(p != q) {
		if(q) { *q = '\0' ; q++ ; } 
		o->designator = my_strdup(p) ;
	} else {
		o->designator = my_strdup("") ;
	}

	/* reading next token: option type, e.g. "float" for a floating number */
	p = strchr(q, '(') ;
	if(!p) return EXIT_SUCCESS ;
	p++ ;
	q = strchr(p, ')') ;
	if(p == q) o->type = BOOLEAN ;
	else {
		if(q) { *q = '\0' ; q++ ; }
		if(strcmp(p, "int") == 0) o->type = INT ;
		else if(strcmp(p, "float") == 0) o->type = DOUBLE ;
		else if(strcmp(p, "file") == 0) o->type = FILENAME ;
		else if(strcmp(p, "string") == 0) o->type = STRING ;
		else if(strcmp(p, "invisible") == 0) o->type = HIDDEN ;
		else Warn("bad option in line %i: %s", i, p) ;
	}

	/* reading default value */
	p = strchr(q, '(') ;
	if(!p) return EXIT_SUCCESS ;
	q = strchr(p, ')') ;
	p++ ;

	if(p != q) {
		if(q) { *q = '\0' ; q++ ; } 
		o->defvalue = my_strdup(p) ;
		o->set = TRUE ;
		switch(o->type){
			case BOOLEAN:
			case INT:
				o->intvalue = atoi(o->defvalue) ;
				break ;
			case DOUBLE:
				o->doublevalue = atof(o->defvalue) ;
				break ;
			case FILENAME:
			default:
				o->strvalue = my_strdup(o->defvalue) ;
				break ;
		}
	}

	/* reading description of the option */
	p = strchr(q, '(') ;
	if(!p) return EXIT_SUCCESS ;
	q = strchr(p, ')') ;
	p++ ;

	if(p != q) {
		if(q) { *q = '\0' ; q++ ; } 
		o->description = my_strdup(p) ;
	}

	/* additional keywords? */
	p = strchr(q, '(') ;
	if(!p) return EXIT_SUCCESS ;
	q = strchr(p, ')') ;
	p++ ;
	if(p != q) {
		if(q) { *q = '\0' ; q++ ; } 
		if(strcmp(p, "required") == 0) {
			o->required = TRUE ;
			o->set = TRUE ;
		}
	}

	return EXIT_SUCCESS ;
}


/* main function parsing the programrc file */
entry_s * parse_rc(FILE *in, int *line) {
	
	char bufor[BUFSIZ], *p, *q, *b ;
		/* the following three variables record the current state of the
		 * reading - that is, in which section of the rcfile we are now */
	int reading_options = FALSE, 
		reading_description = FALSE,

		section = 0,

		/* keeping track of start(program) and end(program) */
		program_found = FALSE,
		program_end_found = FALSE;
	entry_s *result ;

	result = NULL ;

	/* skipping to the first start(program) entry */
	while( !program_found && (fgets(bufor, BUFSIZ, in)) != NULL) {
		(*line)++ ;

		if( (p = strstr(bufor, "start(program)")) == bufor) {
			result = malloc(sizeof(*result)) ;

			p += strlen("start(program)") ;
			if(parse_program(p, result) == EXIT_FAILURE) {
				fprintf(stderr, "wrong entry in line %i\n", *line) ;
				free(result) ;
				return NULL ;
			}
			program_found = TRUE ;
		}
	}


	if(!program_found) return NULL ;

	while(!program_end_found && (fgets(bufor, BUFSIZ, in)) != NULL) {
		(*line)++ ;

		/* skipping comments and empty lines */
		if(*bufor == '\n' || *bufor == '#') continue ;
		b = bufor ;
		while(isspace(*b)) b++ ;

		section = 0 ;
		if(strstr(b, "start") == b) section = 1 ;
		if(strstr(b, "end") == b) section = 2 ;
		
		/* read one of the keywords -- 'start' or 'end' */
		if(section) {
			p = strchr(bufor, '(') ;
			q = strchr(p, ')') ;
			if(!p || !q) {
				fprintf(stderr, "error in configuration file line %i\n", *line) ;
				continue ;
			}
			p++ ;
			*q ='\0' ;
			q++ ;
		} 

		/* a 'normal line' */
		if(section == 0) {

			if(reading_options) 
				if(parse_option(bufor, result, *line)) 
					fprintf(stderr, "error parsing option in line %i\n", *line) ;

			if(reading_description) 
				if(parse_description(bufor, result)) 
					fprintf(stderr, "error parsing a description line %i\n", *line) ;
		}

		/* read a 'start(something)' section */
		if(section == 1) {

			/* section parameter is 'options' */
			if(strcmp(p, "options") == 0 && !reading_description) {
				if(reading_options) {
					fprintf(stderr, "repetition of start(options) on line %i\n", *line) ;
				}
				reading_options = TRUE ;
			}

			/* section parameter is 'description' */
			if(strcmp(p, "description") == 0 && !reading_description) {
				if(reading_options) {
					fprintf(stderr, "warn: start(options) unfinished on line %i\n", *line) ;
					reading_options = FALSE ;
				}
				reading_description = TRUE ;
			}
					
		} /* end of the 'if(section == 1)' statement */

		/* bufor starts with an end() statement */
		if(section == 2) {

			/* end of the description */
			if(strcmp(p, "description") == 0) {
				if(!reading_description) {
					fprintf(stderr, 
						"warn: orphan end(description) section at line %i\n", *line) ;
				}
				reading_description = FALSE ;
			}

			/* end of a programs section */
			if(strcmp(p, "program") == 0 && !reading_description) {
				if(reading_options) {
					fprintf(stderr, 
						"warn: orphan start(options) section in line %i\n", *line) ;
				}
				reading_description = FALSE ;
				reading_options = FALSE ;
				program_end_found = TRUE ;
			}

			/* end of an options section */
			if(strcmp(p, "options") == 0 && !reading_description) {
				if(!reading_options)
					fprintf(stderr, 
						"warn: orphan end(options) section in line %i\n", *line) ;
				reading_options = FALSE ;
			}
			
		} /* end of the if(section == 2) statement */


	} /* end of the main 'while' loop */

	if(!program_end_found) 
		fprintf(stderr, "Warning: no end(program) statement found\n") ;

	return result ;
}
